from typing import Any, Union

from great_expectations.checkpoint.actions import (
    ExpectationSuiteValidationResult,
    GXCloudIdentifier,
    ValidationAction,
    ValidationResultIdentifier,
)
from great_expectations.validator.validator import Validator
from loguru import logger
from oddrn_generator.generators import GreatExpectationsGenerator

from odd_great_expectations.client import Client
from odd_great_expectations.dataset.get_dataset import get_datasets
from odd_great_expectations.mapper import MapValidationResult


class ODDAction(ValidationAction):
    def __init__(
        self,
        data_context,
        data_source_name: str,
        platform_host: str = None,
        platform_token: str = None,
    ):
        self._odd_client = Client(platform_host, platform_token)
        self._data_source_name = data_source_name
        self._host = "local"

        super().__init__(data_context)

    def _run(
        self,
        validation_result_suite: ExpectationSuiteValidationResult,
        validation_result_suite_identifier: Union[
            ValidationResultIdentifier, GXCloudIdentifier
        ],
        data_asset: Union[Validator, Any],
        expectation_suite_identifier=None,
        checkpoint_identifier=None,
        payload=None,
    ):
        try:
            logger.info("Start collecting metadata")
            client = self._odd_client
            expectation_suite = data_asset.expectation_suite
            suite_name = expectation_suite.expectation_suite_name
            generator = GreatExpectationsGenerator(
                host_settings=self._host, suites=suite_name
            )

            docs_link = None
            if payload:
                if data_docs := payload.get("update_data_docs"):
                    docs_link = data_docs.get("local_site")

            client.ingest_data_source(
                data_source_oddrn=generator.get_data_source_oddrn(),
                name=self._data_source_name,
            )

            datasets = get_datasets(data_asset.execution_engine)

            data_entity_list = MapValidationResult(
                suite_result=validation_result_suite,
                suite_result_identifier=validation_result_suite_identifier,
                generator=generator,
                datasets=datasets,
                docs_link=docs_link,
            ).map()

            client.ingest_data_entities(data_entity_list)

            logger.success(
                f"Metadata successfully loaded to Platform. Ingested {len(data_entity_list.items)} entities"
            )
        except Exception as e:
            logger.error(f"Error during collecting metadata. {e}")



from pathlib import Path

from great_expectations.execution_engine import PandasExecutionEngine
from oddrn_generator.generators import FilesystemGenerator


def get_file_dataset(engine: PandasExecutionEngine) -> list[str]:
    batch_spec = engine.batch_manager.active_batch.batch_spec
    batch_path = str(Path(batch_spec["path"]).resolve())
    generator = FilesystemGenerator(host_settings="local", path=batch_path)
    return [generator.get_oddrn_by_path("path")]



class CreateDataSourceError(Exception):
    def __init__(
        self, data_source_name: str, data_source_oddrn: str, message: str
    ) -> None:
        super().__init__(
            f"Couldn't register datasource {data_source_name} with oddrn {data_source_oddrn} \n"
            f"Message: {message}"
        )


class IngestionEntitiesError(Exception):
    def __init__(self, data_source_oddrn: str, message: str) -> None:
        super().__init__(
            f"Couldn't ingest entities for {data_source_oddrn=}.\n"
            f"Message: {message}"
        )



import datetime
from typing import Any, Optional, Tuple

from great_expectations.checkpoint.actions import (
    ExpectationSuiteValidationResult, ValidationResultIdentifier)
from great_expectations.core import ExpectationConfiguration
from great_expectations.expectations.registry import get_expectation_impl
from great_expectations.validator.validator import ExpectationValidationResult
from odd_models.models import (DataEntity, DataEntityList, DataEntityType,
                               DataQualityTest, DataQualityTestExpectation,
                               DataQualityTestRun, LinkedUrl, QualityRunStatus)
from oddrn_generator.generators import GreatExpectationsGenerator


class MapValidationResult:
    def __init__(
        self,
        suite_result: ExpectationSuiteValidationResult,
        suite_result_identifier: ValidationResultIdentifier,
        generator: GreatExpectationsGenerator,
        datasets: list[str],
        docs_link: Optional[str],
    ) -> None:
        self._generator = generator
        self._datasets = datasets
        self._result = suite_result
        self._result_identifier = suite_result_identifier
        self._docs_link = docs_link

    def map(self) -> DataEntityList:
        data_entities = []
        for result in self._result.results:
            data_entities.extend(self._map_result(result))

        return DataEntityList(
            data_source_oddrn=self._generator.get_data_source_oddrn(),
            items=data_entities,
        )

    def _map_result(
        self, validation_result: ExpectationValidationResult
    ) -> tuple[DataEntity, DataEntity]:
        run_id = self._result_identifier.run_id
        status, status_reason = self.get_status(validation_result)

        job = self.map_config(validation_result.expectation_config)
        oddrn = self._generator.get_oddrn_by_path("runs", run_id.run_name)

        run = DataEntity(
            oddrn=oddrn,
            name=f"{job.name}:{run_id.run_name}",
            type=DataEntityType.JOB_RUN,
            data_quality_test_run=DataQualityTestRun(
                data_quality_test_oddrn=job.oddrn,
                start_time=run_id.run_time.replace(tzinfo=None).astimezone(),
                end_time=datetime.datetime.now().astimezone(),
                status_reason=status_reason,
                status=status,
            ),
        )
        return job, run

    def get_status(
        self, validation_result: ExpectationValidationResult
    ) -> Tuple[QualityRunStatus, str]:
        status = QualityRunStatus.SUCCESS
        status_reason = None

        if not validation_result.success:
            status = QualityRunStatus.FAILED

            if unexpected := validation_result.result.get("partial_unexpected_list"):
                status_reason = (
                    f"Unexpected values {str(unexpected)}" if unexpected else None
                )

        return status, status_reason

    def map_config(self, config: ExpectationConfiguration) -> DataEntity:
        original_type = config.expectation_type
        unique_path = uniq_name(config)

        oddrn = self._generator.get_oddrn_by_path("types", unique_path)
        suite_name = (
            self._result_identifier.expectation_suite_identifier.expectation_suite_name
        )

        linked_url_list = None
        if self._docs_link and not self._docs_link.startswith("file://"):
            linked_url_list = [LinkedUrl(name="docs", url=self._docs_link)]

        dqt = DataQualityTest(
            suite_name=suite_name,
            dataset_list=self._datasets,
            expectation=DataQualityTestExpectation(
                type=original_type, **flat_kwargs(config.kwargs)
            ),
            linked_url_list=linked_url_list,
        )

        return DataEntity(
            oddrn=oddrn,
            name=f"{suite_name}:{original_type}",
            type=DataEntityType.JOB,
            data_quality_test=dqt,
        )


def flat_kwargs(kwargs: dict[str, Any]):
    for k, v in kwargs.items():
        if isinstance(v, (list, set, tuple)):
            kwargs[k] = str(list(v))
        elif isinstance(v, dict):
            kwargs[k] = str(v)
    return kwargs


def uniq_name(config: ExpectationConfiguration):
    result = config.expectation_type
    impl = get_expectation_impl(config.expectation_type)

    result += f":{config.kwargs.get('batch_id')}"
    for key in impl.args_keys:
        if value := config.kwargs.get(key):
            if isinstance(value, (list, set, tuple)):
                value = "|".join(value)
            result += f":{key}:{value}"
    return result



from typing import List

from great_expectations.execution_engine import (ExecutionEngine,
                                                 PandasExecutionEngine)
from great_expectations.execution_engine.sqlalchemy_execution_engine import \
    SqlAlchemyExecutionEngine

from odd_great_expectations.dataset.file import get_file_dataset
from odd_great_expectations.dataset.sql_table import get_sql_table_dataset


def get_datasets(execution_engine: ExecutionEngine) -> List[str]:
    datasets: List[str] = []
    if isinstance(execution_engine, PandasExecutionEngine):
        datasets = get_file_dataset(execution_engine)
    elif isinstance(execution_engine, SqlAlchemyExecutionEngine):
        datasets = get_sql_table_dataset(execution_engine)

    return datasets



from typing import List

from great_expectations.execution_engine.sqlalchemy_execution_engine import (
    SqlAlchemyBatchData, SqlAlchemyExecutionEngine)
from oddrn_generator import PostgresqlGenerator
from sqlalchemy.engine import Engine


def postgres_dataset(engine: Engine, batch_data: SqlAlchemyBatchData) -> str:
    generator_params = {
        "host_settings": engine.url.host,
        "databases": engine.url.database,
        "schemas": batch_data.source_schema_name or "public",
        "tables": batch_data.source_table_name,
    }

    generator = PostgresqlGenerator(**generator_params)
    return generator.get_oddrn_by_path("tables")


def get_sql_table_dataset(exec_engine: SqlAlchemyExecutionEngine) -> List[str]:
    dialect_name = exec_engine.dialect_name
    engine = exec_engine.engine
    batch_data = exec_engine.batch_manager.active_batch_data

    if dialect_name == "postgresql":
        return [postgres_dataset(engine, batch_data)]

    raise ValueError(f"Unknown {dialect_name=}")



import os

from odd_models.api_client.open_data_discovery_ingestion_api import ODDApiClient
from odd_models.models import DataEntityList, DataSource, DataSourceList
from requests import HTTPError

from odd_great_expectations.errors import CreateDataSourceError, IngestionEntitiesError


class Client:
    def __init__(self, host: str, token: str) -> None:
        host = host or os.getenv("ODD_PLATFORM_HOST", None)
        token = token or os.getenv("ODD_PLATFORM_TOKEN", None)

        if host is None:
            raise ValueError(
                "Couldn't create HttpClient. Host was not set. Possible solutions:\n"
                "Set env variable :export ODD_PLATFORM_HOST=http://localhost:8080"
            )

        if token is None:
            raise ValueError(
                "Couldn't create HttpClient. Token was not set. Possible solutions:\n"
                "1. Create token using odd-cli\n"
                "2. Create collector token using platform UI\n"
                "Set env variable :export ODD_PLATFORM_TOKEN=<TOKEN_FROM_STEPS_ABOVE>"
            )

        self._client = ODDApiClient(host)
        self._token = token
        self._headers = {"Authorization": f"Bearer {token}"}

    def ingest_data_source(self, data_source_oddrn: str, name: str) -> None:
        response = self._client.create_data_source(
            DataSourceList(items=[DataSource(oddrn=data_source_oddrn, name=name)]),
            headers=self._headers,
        )

        try:
            response.raise_for_status()
        except HTTPError as e:
            message = e.response.json().get("message")
            raise CreateDataSourceError(name, data_source_oddrn, message) from e

    def ingest_data_entities(self, data_entities: DataEntityList) -> None:
        response = self._client.post_data_entity_list(
            data_entities, headers=self._headers
        )

        try:
            response.raise_for_status()
        except HTTPError as e:
            message = e.response.json().get("message")
            raise IngestionEntitiesError(data_entities.data_source_oddrn, message)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['odd_great_expectations', 'odd_great_expectations.dataset']

package_data = \
{'': ['*']}

install_requires = \
['funcy>=1.17,<2.0',
 'great-expectations>=0.15.44,<0.16.0',
 'loguru>=0.6.0,<0.7.0',
 'odd-models>=2.0.10,<3.0.0',
 'oddrn-generator==0.1.62',
 'psycopg2>=2.9.5,<3.0.0',
 'sqlalchemy>=1.4.46,<2.0.0']

setup_kwargs = {
    'name': 'odd-great-expectations',
    'version': '0.1.5',
    'description': 'OpenDataDiscovery Action for Great Expectations',
    'long_description': "## OpenDataDiscovery Great Expectations metadata collecting.\n[![PyPI version](https://badge.fury.io/py/odd-great-expectations.svg)](https://badge.fury.io/py/odd-great-expectations)\n\n## Supporting\n| Feature                     | Supporting |\n| --------------------------- | ---------- |\n| V3 API +                    | +          |\n| SqlAlchemyEngine            | +          |\n| PandasEngine                | +          |\n| Great Expectations V2 API - | -          |\n| Cloud Solution              | -          |\n\n\n`odd_great_expectation.action.ODDAction`\nIs a class derived from `ValidationAction` and can be used in checkpoint actions lists.\n\n## How to:\n\n### Install odd-great-expectations package\n```bash\npip install odd-great-expectations\n```\n\n### Add action to checkpoint:\n```yaml\nname: <CHECKPOINT_NAME>\nconfig_version: 1.0\ntemplate_name:\nmodule_name: great_expectations.checkpoint\nclass_name: Checkpoint\nrun_name_template: '%Y%m%d-%H%M%S-my-run-name-template'\nexpectation_suite_name:\nbatch_request: {}\naction_list:\n  # other actions\n  - name: store_metadata_to_odd \n    action:\n      module_name: odd_great_expectations.action\n      class_name: ODDAction\n      platform_host: <PLATFORM_HOST> # OpenDataDiscovery platform, i.e. http://localhost:8080\n      platform_token: <PLATFORM_TOKEN> # OpenDataDiscovery token\n      data_source_name: <DATA_SOURCE_NAME> # Unique name for data source, i.e. local_qa_test\nevaluation_parameters: {}\n```\n\n### Run checkpoint\n```bash\ngreat_expectations checkpoint run <CHECKPOINT_NAME> \n```\n### Check result\nCheck results on <PLATFORM_HOST> UI.\n\n\n",
    'author': 'Pavel Makarichev',
    'author_email': 'vixtir90@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
